<?php

namespace Pinpoint\PackingSlips\Model\Pdf;

use Magento\Framework\App\Filesystem\DirectoryList;

/**
 * Class GenerateSlip
 * @package Pinpoint\PackingSlips\Model\Pdf
 */
class GenerateSlip extends \Magento\Framework\Model\AbstractModel
{
    /**
     * @var \Dompdf\Dompdf
     */
    protected $domPdf;

    /**
     * @var \Magento\Sales\Api\OrderRepositoryInterface
     */
    protected $orderRepository;

    /**
     * @var \Magento\Catalog\Api\ProductRepositoryInterfaceFactory
     */
    protected $productRepository;

    /**
     * @var \Magento\Framework\Pricing\Helper\Data
     */
    protected $currencyHelper;

    /**
     * @var \Magento\Framework\View\LayoutFactory
     */
    protected $layoutFactory;

    /**
     * @var \Magento\Framework\Filesystem\DirectoryList
     */
    protected $directoryList;

    /**
     * @var string
     */
    protected $fontFolder;

    /**
     * @var string
     */
    protected $fontCacheFolder;

    /**
     * @var \Magento\Catalog\Model\ResourceModel\Product
     */
    protected $productModel;

    /**
     * @var \Magento\Catalog\Model\Product\Attribute\Repository
     */
    protected $productAttributeRepository;

    /**
     * @var array
     */
    protected $machinesRequired;


    /**
     * GenerateSlip constructor.
     * @param \Magento\Catalog\Model\ResourceModel\Product $productResourceModel
     * @param \Magento\Catalog\Model\Product\Attribute\Repository $productAttributeRepository
     * @param \Magento\Sales\Api\OrderRepositoryInterface $orderRepository
     * @param \Magento\Framework\Pricing\Helper\Data $currencyHelper
     * @param \Magento\Catalog\Api\ProductRepositoryInterfaceFactory $productRepositoryFactory
     * @param \Magento\Framework\Filesystem\DirectoryList $dir
     * @param \Magento\Framework\View\LayoutFactory $layoutFactory
     * @param \Magento\Framework\Filesystem $filesystem
     * @param \Magento\Framework\Model\Context $context
     * @param \Magento\Framework\Registry $registry
     * @param \Magento\Framework\Model\ResourceModel\AbstractResource|null $resource
     * @param \Magento\Framework\Data\Collection\AbstractDb|null $resourceCollection
     * @param array $data
     */
    public function __construct(
        \Magento\Catalog\Model\ResourceModel\Product $productResourceModel,
        \Magento\Catalog\Model\Product\Attribute\Repository $productAttributeRepository,
        \Magento\Sales\Api\OrderRepositoryInterface $orderRepository,
        \Magento\Framework\Pricing\Helper\Data $currencyHelper,
        \Magento\Catalog\Api\ProductRepositoryInterfaceFactory $productRepositoryFactory,
        \Magento\Framework\Filesystem\DirectoryList $dir,
        \Magento\Framework\View\LayoutFactory $layoutFactory,
        \Magento\Framework\Filesystem $filesystem,
        \Magento\Framework\Model\Context $context,
        \Magento\Framework\Registry $registry,
        \Magento\Framework\Model\ResourceModel\AbstractResource $resource = null,
        \Magento\Framework\Data\Collection\AbstractDb $resourceCollection = null,
        array $data = []
    ) {
        $mediaDir = DIRECTORY_SEPARATOR . trim($filesystem->getDirectoryRead(DirectoryList::MEDIA)->getAbsolutePath(), '/');
        $fontFolder = $mediaDir . DIRECTORY_SEPARATOR . 'dompdf_fonts';
        $fontCacheFolder = $mediaDir . DIRECTORY_SEPARATOR . 'dompdf_fonts_cache';

        if (!is_dir($fontFolder)) {
            mkdir($fontFolder, 0777, true);
        }

        if (!is_dir($fontCacheFolder)) {
            mkdir($fontCacheFolder, 0777, true);
        }

        $this->fontCacheFolder = $fontCacheFolder;
        $this->fontFolder = $fontFolder;
        $this->directoryList = $dir;
        $this->orderRepository = $orderRepository;
        $this->productRepository = $productRepositoryFactory;
        $this->currencyHelper = $currencyHelper;
        $this->layoutFactory = $layoutFactory;
        $this->productResourceModel = $productResourceModel;
        $this->productAttributeRepository = $productAttributeRepository;

        $this->machinesRequired = $this->getRequiredMachines();
        $this->pppItemType = $this->getPppItemType();

        parent::__construct($context, $registry, $resource, $resourceCollection, $data);
    }

    /**
     * @return array
     * @throws \Magento\Framework\Exception\NoSuchEntityException
     */
    protected function getRequiredMachines()
    {

        $requiredMachineValues = [];
        $requiredMachines = $this->productAttributeRepository->get('required_machines')->getOptions();
        $values = array();

        foreach ($requiredMachines as $requiredMachine) {
            $requiredMachineValues[$requiredMachine->getValue()] = $requiredMachine->getLabel();  // Label
        }

        return $requiredMachineValues;

    }

    protected function getPppItemType()
    {
        $ppp_item_typeValues = [];
        $ppp_item_types = $this->productAttributeRepository->get('ppp_item_type')->getOptions();
        $values = array();

        foreach ($ppp_item_types as $ppp_item_type) {
            $ppp_item_typeValues[$ppp_item_type->getValue()] = $ppp_item_type->getLabel();  // Label
        }

        return $ppp_item_typeValues;


    }

    /**
     * @param $orderId
     * @throws \Exception
     */
    public function generateFromApi($orderId)
    {
        $this->_appState->emulateAreaCode('adminhtml', [$this, "generateSlip"], [$orderId, true]);
    }

    /**
     * @param $orderId
     * @param bool $save
     * @param bool $returnSort
     * @return array|\Dompdf\Dompdf
     */
    public function generateSlip($orderId, $save = false, $returnSort = false)
    {

        try {
            $order = $this->orderRepository->get($orderId);
        } catch (\Exception $ex) {
            return null;
        }

        $pdfData = [
            'order' => [
                'id' => $orderId,
                'increment_id' => $order->getIncrementId(),
                'date_purchased' => $order->getCreatedAt(),
                'payment_method' => $order->getPayment()->getData('additional_information')['method_title'],
                'discount' => $order->getDiscountAmount(),
                'barcode' => $orderId,
                'delivery_method' => strtolower($order->getShippingDescription()),
                'has_logo' => false
            ],
            'customer' => [
                'customer_name' => sprintf('%s %s', $order->getShippingAddress()->getFirstname(), $order->getShippingAddress()->getLastname()),
                'billing_address' => $order->getBillingAddress(),
                'delivery_address' => $order->getShippingAddress(),
            ],
            'order_items' => $order->getItems()
        ];

        $requiredMachines = [];

        /** @var \Magento\Sales\Api\Data\OrderItemInterface $item */
        foreach ($pdfData['order_items'] as $item) {
            $requiredMachines[] = $productMachines = $this->machinesRequired[$this->productResourceModel->getAttributeRawValue($item->getProductId(), 'required_machines', 0)];
        }

        //sort items to put personalised options last
        usort($pdfData['order_items'], function ($a, $b) use ($pdfData)
        {
            $productAOptions = $a->getProductOptions();
            $productBOptions = $b->getProductOptions();

            if ($productAOptions && $productBOptions) {
                if (isset($productAOptions['options']) && isset($productBOptions['options'])) {


                    $machineA = $this->machinesRequired[$this->productResourceModel->getAttributeRawValue($a->getProductId(), 'required_machines', 0)];
                    $machineB = $this->machinesRequired[$this->productResourceModel->getAttributeRawValue($b->getProductId(), 'required_machines', 0)];

                    if ($machineA && $machineB) {
                        if ($machineA == $machineB) {
                            return 0;
                        } else {
                            return 1;
                        }
                    }

                    return 0;
                } else if (isset($productAOptions['options']) && !isset($productBOptions['options'])) {
                    return 1;
                } else {
                    return -1;
                }
            } else if ($productAOptions && !$productBOptions) {
                return 1;
            } else {
                return -1;
            }
        });

        $pdfData['order']['required_machines'] = rtrim(implode(',', array_filter($requiredMachines, 'strlen')), ',');
        $pdfData['order']['id'] = str_pad($pdfData['order']['id'], 9, 0, STR_PAD_LEFT);

        foreach ($pdfData['order_items'] as $item) {
            if (!isset($item->getProductOptions()['options'])) {
                $pdfData['product_pages']['simple'][] = $item;
            } else {
                foreach ($item->getProductOptions()['options'] as $option) {
                    if ($option['option_type'] === 'file') {
                        $pdfData['order']['has_logo'] = true;
                        break;
                    }
                }

                $machine = $this->machinesRequired[$this->productResourceModel->getAttributeRawValue($item->getProductId(), 'required_machines', 0)];

                if ($machine) {
                    $pdfData['product_pages'][$machine][] = $item;
                } else {
                    $pdfData['product_pages']['na'][] = $item;
                }
            }
        }

        foreach ($pdfData['product_pages'] as $prodType => $products) {
            if ($prodType !== 'simple') {
                usort($pdfData['product_pages'][$prodType], function ($a, $b) use ($pdfData) {

                    $itemTypeA = $this->pppItemType[$this->productResourceModel->getAttributeRawValue($a->getProductId(), 'ppp_item_type', 0)];
                    $itemTypeB = $this->pppItemType[$this->productResourceModel->getAttributeRawValue($b->getProductId(), 'ppp_item_type', 0)];

                    if ($itemTypeA === $itemTypeB) {
                        return 0;
                    } else {
                        return 1;
                    }
                });
            }
        }

        $counter = [];
        foreach ($pdfData['product_pages'] as $machine => $product_page) {
            if (!isset($counter[$machine])) {
                $counter[$machine] = 0;
            }

            $counter[$machine] += 1;
        }

        if (count($counter) > 1) {
            $machineSort = 'Split Machine Orders';
        } else {
            $machineSort = array_keys($counter, max($counter))[0];
        }

        $slip = $this->getPdfTemplate($pdfData);

        if ($save) {
            $pdfOutput = $slip->output();
            $outFolder = $this->directoryList->getPath('var') . DIRECTORY_SEPARATOR .
                'export' . DIRECTORY_SEPARATOR .
                'slips' . DIRECTORY_SEPARATOR;

            if (!is_dir($outFolder)) {
                mkdir($outFolder);
            }

            \file_put_contents(
                $outFolder .
                "packing-slip-{$order->getId()}.pdf",
                $pdfOutput
            );
        }

        if (!$save) {
            try {
                $order->setState('printed')->setStatus('printed');
                $this->orderRepository->save($order);
            } catch (\Exception $exception) {
                unset($exception);
            }
        }

        $this->domPdf = null;

        if (!$returnSort) {
            return $slip;
        } else {
            return [$slip, $machineSort, $order];
        }
    }

    /**
     * @param $text
     * @return \Dompdf\Dompdf
     */
    public function createSplitterPage($text)
    {
        $pdf = new \Dompdf\Dompdf([
            'enable_remote' => true,
            'DOMPDF_ENABLE_CSS_FLOAT' => true,
            'defaultFont' => 'Helvetica',
            'fontDir' => $this->fontFolder,
            'fontCache' => $this->fontCacheFolder
        ]);;

        $pdf->loadHtml("<h1>Machine: $text</h1>");
        $pdf->render();

        return $pdf;
    }

    /**
     * @param $params
     * @return mixed
     */
    protected function getPdfTemplate($params)
    {
        return $this->layoutFactory->create()
            ->createBlock('Pinpoint\PackingSlips\Block\Adminhtml\Pdf\PackingSlip')
            ->setData($params)
            ->toHtml()
            ;
    }
}
